from typing import Optional, List
from pykeadhcp.models.generic.client_class import ClientClass
from pykeadhcp.models.generic.option_def import OptionDef


class ClientClass4(ClientClass):
    option_def: Optional[List[OptionDef]]
    next_server: Optional[str]
    server_hostname: Optional[str]
    boot_file_name: Optional[str]



from typing import Optional, List
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.option_data import OptionData


class Pool(KeaBaseModel):
    pool: str
    option_data: Optional[List[OptionData]]
    client_class: Optional[str]
    require_client_classes: Optional[str]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]



from typing import Optional, List
from pykeadhcp.models.generic.client_class import ClientClass


class ClientClass6(ClientClass):
    preferred_lifetime: Optional[int]
    min_preferred_lifetime: Optional[int]
    max_preferred_lifetime: Optional[int]



from typing import Optional
from pykeadhcp.models.generic.config import CommonConfig
from pykeadhcp.models.enums import NCRProtocolEnum, NCRFormatEnum


class DhcpDdns(CommonConfig):
    enable_updates: bool
    server_ip: str
    server_port: int
    sender_ip: str
    sender_port: int
    max_queue_size: int
    ncr_protocol: NCRProtocolEnum
    ncr_format: NCRFormatEnum
    dep_override_no_update: Optional[bool]
    dep_override_client_update: Optional[bool]
    dep_replace_client_name: Optional[str]
    dep_generated_prefix: Optional[str]
    dep_qualifying_suffix: Optional[str]
    dep_hostname_char_set: Optional[str]
    dep_hostname_char_replacement: Optional[str]


"""
"enable-updates": false,
"max-queue-size": 1024,
"ncr-format": "JSON",
"ncr-protocol": "UDP",
"sender-ip": "0.0.0.0",
"sender-port": 0,
"server-ip": "127.0.0.1",
"server-port": 53001
"""



from typing import Optional
from pykeadhcp.models.generic.config import CommonConfig


class OptionDef(CommonConfig):
    name: str
    code: Optional[int]
    type: Optional[str]
    record_types: Optional[str]
    space: Optional[str]
    encapsulate: Optional[str]
    array: Optional[bool]



from pykeadhcp.parsers.generic import GenericParser
from pykeadhcp.models.dhcp6.config import Dhcp6DaemonConfig
from pykeadhcp.models.dhcp6.subnet import Subnet6


class Dhcp6Parser(GenericParser):
    """Parser for the ISC Kea Dhcp6 configuration file. This should ideally
    be used with the cached config stored in the Daemon class like this:

    parser = Dhcp6Parser(config=server.dhcp6.cached_config)
    """

    def __init__(self, config: dict):
        self.config = Dhcp6DaemonConfig.parse_obj(config["Dhcp6"])



from typing import Optional, List
from pykeadhcp.models.generic.daemon import CommonDhcpDaemonConfig
from pykeadhcp.models.dhcp4.client_class import ClientClass4
from pykeadhcp.models.dhcp4.shared_network import SharedNetwork4
from pykeadhcp.models.dhcp4.subnet import Subnet4
from pykeadhcp.models.dhcp4.reservation import Reservation4


class Dhcp4DaemonConfig(CommonDhcpDaemonConfig):
    client_classes: Optional[List[ClientClass4]]
    shared_networks: Optional[List[SharedNetwork4]]
    reservations: Optional[List[Reservation4]]
    subnet4: Optional[List[Subnet4]]
    echo_client_id: Optional[bool]
    match_client_id: Optional[bool]
    authoritative: Optional[bool]
    next_server: Optional[str]
    server_hostname: Optional[str]
    boot_file_name: Optional[str]



from typing import List


class GenericParser:
    """A Parser does not interact with the ISC Kea Daemon APIs but essentially builds
    similar functionality for a local cached config file. If you do not pay for the
    premium hooks to expose the different commands like subnet4-add or network4-del then
    the parsers job is to provide similar functionality for the provided daemon configuration
    and then you should use the relevant service `config-set` functionality which is implemented
    for each service in pykeadhcp.

    The recommended option is to use the functions within the Daemon class (eg. server.dhcp4) and to
    use these parsers as a last resort as tests are not currently performed as extensively vs the API
    functionality"""

    def __init__(self, config: dict):
        self.config = config



from typing import Optional, List, Union
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.config import CommonDhcpConfig
from pykeadhcp.models.generic.hook import Hook
from pykeadhcp.models.generic.logger import Logger
from pykeadhcp.models.generic.database import Database
from pykeadhcp.models.generic.option_def import OptionDef
from pykeadhcp.models.generic.control_socket import ControlSocket
from pykeadhcp.models.generic.dhcp_queue_control import DHCPQueueControl
from pykeadhcp.models.generic.dhcp_ddns import DhcpDdns
from pykeadhcp.models.generic.sanity_check import SanityCheck
from pykeadhcp.models.generic.multi_threading import MultiThreading
from pykeadhcp.models.enums import (
    DHCPSocketTypeEnum,
    OutboundInterfaceEnum,
    HostReservationIdentifierEnum,
)


class CommonDaemonConfig(CommonDhcpConfig):
    hooks_libraries: Optional[List[Hook]]
    loggers: Optional[List[Logger]]


class InterfaceListConfig(KeaBaseModel):
    interfaces: List[str]
    dhcp_socket_type: Optional[DHCPSocketTypeEnum]
    outbound_interface: Optional[OutboundInterfaceEnum]
    re_detect: Optional[bool]
    service_sockets_require_all: Optional[bool]
    service_sockets_retry_wait_time: Optional[int]
    service_sockets_max_retries: Optional[int]


class CommonDhcpDaemonConfig(CommonDaemonConfig):
    interfaces_config: InterfaceListConfig
    lease_database: Optional[Database]
    hosts_database: Optional[Database]
    hosts_databases: Optional[List[Database]]
    host_reservation_identifiers: Optional[List[HostReservationIdentifierEnum]]
    option_def: Optional[List[OptionDef]]
    expired_leases_processing: Optional[dict]
    dhcp4o6_port: Optional[int]
    control_socket: Optional[ControlSocket]
    dhcp_queue_control: Optional[DHCPQueueControl]
    dhcp_ddns: Optional[DhcpDdns]
    sanity_checks: Optional[SanityCheck]
    config_control: Optional[dict]
    server_tag: Optional[str]
    hostname_char_set: Optional[str]
    hostname_char_replacement: Optional[str]
    statistic_default_sample_count: Optional[int]
    statistic_default_sample_age: Optional[int]
    multi_threading: Optional[MultiThreading]
    early_global_reservations_lookup: Optional[bool]
    ip_reservations_unique: Optional[bool]
    reservations_lookup_first: Optional[bool]
    compatibility: Optional[dict]
    parked_packet_limit: Optional[int]
    decline_probation_period: Optional[int]



from typing import Optional
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.enums import DatabaseTypeEnum, DatabaseOnFailEnum


class Database(KeaBaseModel):
    type: DatabaseTypeEnum
    user: Optional[str]
    password: Optional[str]
    host: Optional[str]
    port: Optional[int]
    name: Optional[str]
    persist: Optional[bool]
    lfc_interval: Optional[int]
    readonly: Optional[bool]
    connect_timeout: Optional[int]
    max_reconnect_tries: Optional[int]
    on_fail: Optional[DatabaseOnFailEnum]
    max_row_errors: Optional[int]
    trust_anchor: Optional[str]
    cert_file: Optional[str]
    key_file: Optional[str]
    cipher_list: Optional[str]
    unknown_map_entry: Optional[str]



from typing import Any, List, Optional, TYPE_CHECKING
from pydantic import BaseModel

from pykeadhcp.models.generic.shared_network import SharedNetwork
from pykeadhcp.models.dhcp4.subnet import Subnet4


class SharedNetwork4(SharedNetwork):
    subnet4: Optional[List[Subnet4]]
    match_client_id: Optional[bool]
    authoritative: Optional[bool]
    next_server: Optional[str]
    server_hostname: Optional[str]
    boot_file_name: Optional[str]



from typing import Optional, List
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.option_data import OptionData


class PDPool(KeaBaseModel):
    prefix: str
    prefix_len: int
    delegated_len: int
    option_data: Optional[List[OptionData]]
    client_class: Optional[str]
    require_client_classes: Optional[List[str]]
    excluded_prefix: Optional[str]
    excluded_prefix_len: Optional[int]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]



from typing import Optional, List
from pydantic import BaseModel
from pykeadhcp.models.generic import Sockets


class StatusGet(BaseModel):
    pid: int
    uptime: int
    reload: int
    multi_threading_enabled: Optional[bool]
    sockets: Optional[Sockets]
    high_availability: Optional[List[dict]]

    class Config:
        fields = {
            "multi_threading_enabled": "multi-threading-enabled",
            "high_availability": "high-availability",
        }



from pykeadhcp.parsers.generic import GenericParser
from pykeadhcp.models.dhcp4.config import Dhcp4DaemonConfig
from pykeadhcp.models.dhcp4.shared_network import SharedNetwork4
from pykeadhcp.models.dhcp4.subnet import Subnet4


class Dhcp4Parser(GenericParser):
    """Parser for the ISC Kea Dhcp4 configuration file. This should ideally
    be used with the cached config stored in the Daemon class like this:

    parser = Dhcp4Parser(config=server.dhcp4.cached_config)
    """

    def __init__(self, config: dict):
        self.config = Dhcp4DaemonConfig.parse_obj(config["Dhcp4"])

    def get_shared_network(self, name: str) -> SharedNetwork4:
        """Returns a specific Dhcp4 shared-network

        Args:
            name:       Name of the shared-network
        """
        for shared_network in self.config.shared_networks:
            if shared_network.name == name:
                return shared_network

    def get_subnet(self, id: int) -> Subnet4:
        """Attempts to return the first found subnet based on the subnets id using
        global subnets first then shared-networks last.

        Args:
            id:     ID of the subnet
        """
        for subnet in self.config.subnet4:
            if subnet.id == id:
                return subnet

        for shared_network in self.config.shared_networks:
            for subnet in shared_network.subnet4:
                if subnet.id == id:
                    return subnet

    def get_subnet_by_cidr(self, cidr: str) -> Subnet4:
        """Attempts to return the first found subnet based on the provided cidr
        using global subnets first and then shared-networks last.

        Args:
            cidr:       IPv4 CIDR (eg. 192.0.2.0/24)
        """
        for subnet in self.config.subnet4:
            if subnet.subnet == cidr:
                return subnet

        for shared_network in self.config.shared_networks:
            for subnet in shared_network.subnet4:
                if subnet.subnet == cidr:
                    return subnet



from typing import Optional
from pykeadhcp.models.generic.base import KeaBaseModel


class OptionData(KeaBaseModel):
    data: str
    name: Optional[str]
    code: Optional[int]
    space: Optional[str]
    csv_format: Optional[bool]
    always_send: Optional[bool]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]



from typing import List, Optional
from pykeadhcp.models.generic.subnet import Subnet
from pykeadhcp.models.dhcp6.pd_pool import PDPool
from pykeadhcp.models.dhcp6.reservation import Reservation6


class Subnet6(Subnet):
    preferred_lifetime: Optional[int]
    min_preferred_lifetime: Optional[int]
    max_preferred_lifetime: Optional[int]
    pd_pools: Optional[List[PDPool]]
    interface_id: Optional[str]
    rapid_commit: Optional[bool]
    reservations: Optional[List[Reservation6]]



from typing import Optional
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.config import CommonConfig


class ControlSocket(CommonConfig):
    socket_name: str
    socket_type: str


class ControlSockets(KeaBaseModel):
    dhcp4: Optional[ControlSocket]
    dhcp6: Optional[ControlSocket]
    d2: Optional[ControlSocket]
    unknown_map_entry: Optional[str]



from typing import Optional
from pykeadhcp.models.generic.reservation import Reservation


class Reservation4(Reservation):
    client_id: Optional[str]
    circuit_id: Optional[str]
    ip_address: str
    next_server: Optional[str]
    server_hostname: Optional[str]
    boot_file_name: Optional[str]



from typing import Optional, List
from pykeadhcp.models.generic.daemon import CommonDaemonConfig
from pykeadhcp.models.generic.control_socket import ControlSockets


class CtrlAgentDaemonConfig(CommonDaemonConfig):
    http_host: str
    http_port: int
    trust_anchor: Optional[str]
    cert_file: Optional[str]
    key_file: Optional[str]
    cert_required: Optional[bool]
    control_sockets: Optional[ControlSockets]
    authentication: Optional[dict]



from typing import Optional, List
from pykeadhcp.models.generic.reservation import Reservation


class Reservation6(Reservation):
    ip_addresses: Optional[List[str]]
    prefixes: Optional[List[str]]



from typing import Optional
from pykeadhcp.models.generic.config import CommonConfig
from pykeadhcp.models.generic.option_data import OptionData


class ClientClass(CommonConfig):
    name: str
    test: Optional[str]
    only_if_required: Optional[bool]
    option_data: Optional[OptionData]
    valid_lifetime: Optional[int]
    min_valid_lifetime: Optional[int]
    max_valid_lifetime: Optional[int]



from typing import Optional, List, Union
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.hook import Hook
from pykeadhcp.models.generic.logger import Logger
from pykeadhcp.models.generic.option_data import OptionData
from pykeadhcp.models.enums import ReservationMode, DDNSReplaceClientNameEnum


class CommonConfig(KeaBaseModel):
    store_extended_info: Optional[bool]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]


class CommonDhcpConfig(CommonConfig):
    valid_lifetime: Optional[int]
    min_valid_lifetime: Optional[int]
    max_valid_lifetime: Optional[int]
    renew_timer: Optional[int]
    rebind_timer: Optional[int]
    option_data: Optional[List[OptionData]]
    reservation_mode: Optional[ReservationMode]
    reservations_global: Optional[bool]
    reservations_in_subnet: Optional[bool]
    reservations_out_of_pool: Optional[bool]
    calculate_tee_times: Optional[bool]
    t1_percent: Optional[float]
    t2_percent: Optional[float]
    cache_threshold: Optional[float]
    cache_max_age: Optional[int]
    ddns_send_updates: Optional[bool]
    ddns_override_no_update: Optional[bool]
    ddns_override_client_update: Optional[bool]
    ddns_replace_client_name: Optional[Union[DDNSReplaceClientNameEnum, bool]]
    ddns_generated_prefix: Optional[str]
    ddns_qualifying_suffix: Optional[str]
    ddns_update_on_renew: Optional[bool]
    ddns_use_conflict_resolution: Optional[bool]



from typing import List, Optional
from pykeadhcp.models.generic.subnet import Subnet
from pykeadhcp.models.dhcp4.reservation import Reservation4


class Subnet4(Subnet):
    match_client_id: Optional[bool]
    authoritative: Optional[bool]
    next_server: Optional[str]
    boot_file_name: Optional[str]
    subnet_4o6_interface: Optional[str]
    subnet_4o6_interface_id: Optional[str]
    subnet_4o6_subnet: Optional[str]
    reservations: Optional[List[Reservation4]]

    class Config:
        fields = {
            "subnet_4o6_interface": "4o6-interface",
            "subnet_4o6_interface_id": "4o6-interface-id",
            "subnet_4o6_subnet": "4o6-subnet",
        }



from typing import Optional, List
from pydantic import conint
from pykeadhcp.models.generic.dhcp_common import CommonDHCPParams
from pykeadhcp.models.generic.pool import Pool


class Subnet(CommonDHCPParams):
    id: Optional[conint(gt=0, lt=4294967295)]
    pools: Optional[List[Pool]]
    subnet: str
    hostname_char_set: Optional[str]
    hostname_char_replacement: Optional[str]



import requests
from requests.auth import HTTPBasicAuth
from requests.exceptions import HTTPError
from pathlib import Path
from typing import List

from pykeadhcp.daemons import CtrlAgent, Ddns, Dhcp4, Dhcp6
from pykeadhcp.models.generic import KeaResponse
from pykeadhcp.models.generic.hook import Hook
from pykeadhcp.exceptions import (
    KeaGenericException,
    KeaCommandNotSupportedException,
    KeaObjectNotFoundException,
    KeaServerConflictException,
    KeaUnauthorizedAccessException,
    KeaHookLibraryNotConfiguredException,
)


class Kea:
    """Basic wrapper around requests module for interacting with the
    Kea Management API for the various daemons supported.

    Args:
        ip:                     IP address of the Kea server
        port:                   TCP Port of the Kea Server to access the API
        headers:                Headers to inject in every POST request sent to the API
        use_basic_auth:         Use HTTP Basic Auth if Kea is configured for it
        username:               Username for HTTP Basic Auth
        password:               Password for HTTP Basic Auth
        raise_generic_errors:   If True, it will raise a generic error based on Kea Documentation
            (as per self.RESPONSE_CODES). Set this to False if you want to catch the API endpoint specific
            errors such as `v4 Shared Network not found` vs `Code 3: the requested operation has been completed but the requested resource was not found.
            This status code is returned when a command returns no resources or affects no resources.`
    """

    def __init__(
        self,
        host: str,
        port: int,
        headers: dict = {"Content-Type": "application/json"},
        use_basic_auth: bool = False,
        username: str = "",
        password: str = "",
        raise_generic_errors: bool = False,
    ):
        self.host = host
        self.port = port
        self.headers = headers
        self.use_basic_auth = use_basic_auth
        self.basic_auth = HTTPBasicAuth(username, password)
        self.services = ["dhcp4", "dhcp6", "ddns", None]  # None = Control-Agent Daemon
        self.url = f"{self.host}:{self.port}"
        self.raise_generic_errors = raise_generic_errors
        self.RESPONSE_CODES = {
            1: KeaGenericException,
            2: KeaCommandNotSupportedException,
            3: KeaObjectNotFoundException,
            4: KeaServerConflictException,
        }
        self.hook_library = {}
        self.ctrlagent = CtrlAgent(self)
        self.ddns = Ddns(self)
        self.dhcp4 = Dhcp4(self)
        self.dhcp6 = Dhcp6(self)

    def get_active_hooks(self, hooks: List[dict]) -> List[Hook]:
        """Returns a list of Hook objects

        Args:
            hooks:      List of hook libraries
        """
        hook_libraries = [
            Hook(
                library=hook["library"],
                parameters=hook.get("parameters", {}),
                name=self.get_hook_name(hook["library"]),
            )
            for hook in hooks
        ]
        return hook_libraries

    def get_hook_name(self, hook_path: str) -> str:
        """Returns a friendly name for the hook library similar to Kea Documentation

        eg. if the library path is: /usr/lib/x86_64-linux-gnu/kea/hooks/libdhcp_lease_cmds.so
        then we can simply take the file name, remove the suffix and `libdhcp_` part to get `lease_cmds`.

        Args:
            hook_path:      Hook Library Path
        """
        file = Path(hook_path)
        filename = file.stem.split("libdhcp_")[-1]
        return filename

    def is_hook_enabled(self, hook: str, hook_library: List[Hook]) -> bool:
        """Checks the provided hook library if the hook is configured on the specific Daemon config

        Args:
            hook:           Name of hook (shortname as per documentation) to check
            hook_library:   List of Hooks enabled
        """
        for configured_hook in hook_library:
            if not configured_hook.name:
                continue

            if hook == configured_hook.name:
                return True

        return False

    def post(self, endpoint: str, body: dict, **kwargs) -> KeaResponse:
        """Handles simple POST operation and basic header injection

        Args:
            endpoint:       API Endpoint
            body:           JSON body to send
        """
        url = self.url + endpoint
        response = requests.post(
            url=url,
            json=body,
            headers=self.headers,
            auth=self.basic_auth if self.use_basic_auth else None,
            **kwargs,
        )

        if response.status_code == 401:
            raise KeaUnauthorizedAccessException

        if response.status_code >= 400 and response.status_code <= 500:
            response.raise_for_status()

        data = response.json()
        if not data:
            return None

        data = data[0]  # Kea API returns everything in a list
        result_code = data["result"]
        if self.raise_generic_errors and result_code != 0:
            raise self.RESPONSE_CODES.get(
                result_code, KeaGenericException
            )  # Return Generic Exception if code not found

        return KeaResponse(**data)

    def send_command(
        self, command: str, service: str, required_hook: str = ""
    ) -> KeaResponse:
        """Sends a command to the specific API daemon

        Args:
            command:        Supported command by the daemons API
            service:        Service to send request to
            required_hook:  Precheck if hook library is enabled
        """
        if service and service.lower() not in self.services:
            raise TypeError(
                f"Service {service} is not a supported service. The supported services are {self.services}"
            )

        if required_hook and not self.is_hook_enabled(
            required_hook, self.hook_library[service]
        ):
            raise KeaHookLibraryNotConfiguredException(service, required_hook)

        command_results = self.post(
            endpoint="/",
            body={"command": command, "service": [service] if service else []},
        )
        return command_results

    def send_command_with_arguments(
        self, command: str, service: str, arguments: dict, required_hook: str = ""
    ) -> KeaResponse:
        """Sends a command to the specific API daemon with provided arguments

        Args:
            command:        Supported command by the daemons API
            service:        Service to send request to
            arguments:      Argument parameters to pass to the command/service
            required_hook:  Precheck if hook library is enabled
        """
        if service and service.lower() not in self.services:
            raise TypeError(
                f"Service {service} is not a supported service. The supported services are {self.services}"
            )

        if required_hook and not self.is_hook_enabled(
            required_hook, self.hook_library[service]
        ):
            raise KeaHookLibraryNotConfiguredException(service, required_hook)

        command_results = self.post(
            endpoint="/",
            body={
                "command": command,
                "service": [service] if service else [],
                "arguments": arguments,
            },
        )
        return command_results

    def get_next_available_subnet_id(self, subnet_ids: List[int]) -> int:
        """Returns the next available subnet-id based on a given list of
        existing subnet-ids

        Args:
            subnet_ids:     Existing list of subnet-ids gathered via
                subnet4_list/subnet6_list or manually gathered via parser
        """
        next_id = next(
            i for i, e in enumerate(sorted(subnet_ids) + [None], 1) if i != e
        )
        return next_id



from typing import Optional, List
from pykeadhcp.models.generic.config import CommonDhcpConfig
from pykeadhcp.models.generic.relay import Relay


class CommonDHCPParams(CommonDhcpConfig):
    """Any param shared between v4/v6 Shared Networks and Subnets"""

    interface: Optional[str]
    client_class: Optional[str]
    require_client_classes: Optional[List[str]]
    relay: Optional[Relay]



from enum import Enum


class StatusEnum(str, Enum):
    ready = "ready"
    retrying = "retrying"
    failed = "failed"


class ReservationMode(str, Enum):
    disabled = "disabled"
    out_of_pool = ("out-of-pool",)
    r_global = "global"
    all = "all"


class DDNSReplaceClientNameEnum(str, Enum):
    when_present = "when-present"
    never = "never"
    always = "always"
    when_not_present = "when-not-present"


class Lease6TypeEnum(str, Enum):
    iana = "IA_NA"
    iapd = "IA_PD"


class LoggerLevelEnum(str, Enum):
    critical = "CRITICAL"
    error = "ERROR"
    warning = "WARNING"
    info = "INFO"
    debug = "DEBUG"


class DHCPSocketTypeEnum(str, Enum):
    raw = "raw"
    udp = "udp"


class OutboundInterfaceEnum(str, Enum):
    same_as_inbound = "same-as-inbound"
    use_routing = "use-routing"


class DatabaseTypeEnum(str, Enum):
    memfile = "memfile"
    mysql = "mysql"
    postgresql = "postgresql"


class DatabaseOnFailEnum(str, Enum):
    stop_retry_exit = "stop-retry-exit"
    serve_retry_exit = "serve-retry-exit"
    serve_retry_continue = "serve-retry-continue"


class HostReservationIdentifierEnum(str, Enum):
    duid = "duid"
    hw_address = "hw-address"
    circuit_id = "circuit-id"
    client_id = "client-id"
    flex_id = "flex-id"


class NCRProtocolEnum(str, Enum):
    udp = "UDP"
    tcp = "TCP"


class NCRFormatEnum(str, Enum):
    json = "JSON"


class ServerIdTypeEnum(str, Enum):
    llt = "LLT"
    en = "EN"
    ll = "LL"


class AuthenticationTypeEnum(str, Enum):
    basic = "basic"



from typing import Optional, List
from pydantic import BaseModel
from pykeadhcp.models.enums import StatusEnum


class Sockets(BaseModel):
    errors: Optional[List[str]]
    status: StatusEnum

    class Config:
        use_enum_values = True



from typing import Optional, List
from pykeadhcp.models.generic.shared_network import SharedNetwork
from pykeadhcp.models.dhcp6.subnet import Subnet6


class SharedNetwork6(SharedNetwork):
    subnet6: Optional[List[Subnet6]]
    interface_id: Optional[str]
    min_preferred_lifetime: Optional[int]
    max_preferred_lifetime: Optional[int]
    rapid_commit: Optional[bool]



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from pykeadhcp import Kea

from pykeadhcp.models.generic import KeaResponse, StatusGet
from pykeadhcp.models.dhcp6.lease import Lease6, Lease6Page, Lease6TypeEnum
from pykeadhcp.models.dhcp6.pd_pool import PDPool
from pykeadhcp.models.dhcp6.reservation import Reservation6
from pykeadhcp.models.dhcp6.shared_network import SharedNetwork6
from pykeadhcp.models.dhcp6.subnet import Subnet6
from pykeadhcp.exceptions import (
    KeaSharedNetworkNotFoundException,
    KeaSubnetNotFoundException,
    KeaLeaseNotFoundException,
)


class Dhcp6:
    def __init__(self, api: "Kea"):
        self.service = "dhcp6"
        self.api = api

        # Cache config and hooks
        try:
            self.cached_config = None
            self.refresh_cached_config()
            self.hook_libraries = self.api.get_active_hooks(
                hooks=self.cached_config[self.service.capitalize()]["hooks-libraries"]
            )
            self.api.hook_library[self.service] = self.hook_libraries
        except:
            pass

    def refresh_cached_config(self):
        """Sets the cached_config variable

        This function should be called after any interaction with the API that potentially changes the configuration
        eg. subnet6-add, commands like lease6-add won't need a config refresh to keep the cached config up to date
        """
        config = self.config_get()
        self.cached_config = config.arguments

    def get_next_available_subnet_id(self) -> int:
        """Returns the next available subnet-id for use with Dhcp4 subnets"""
        subnets = self.subnet6_list()
        subnet_ids = [subnet.id for subnet in subnets]
        next_id = self.api.get_next_available_subnet_id(subnet_ids=subnet_ids)
        return next_id

    def build_report(self) -> KeaResponse:
        """Returns list of compilation options that this particular binary was built with

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-build-report
        """
        return self.api.send_command(command="build-report", service=self.service)

    def config_get(self) -> KeaResponse:
        """Retrieves the current configuration used by the server

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-config-get
        """
        return self.api.send_command(command="config-get", service=self.service)

    def config_reload(self) -> KeaResponse:
        """Reloads the last good configuration (configuration file on disk)

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-config-reload
        """
        return self.api.send_command(command="config-reload", service=self.service)

    def config_set(self, config: dict) -> KeaResponse:
        """Replace the current server configuration with the provided configuration

        Args:
            config:     Configuration to set

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-set
        """
        return self.api.send_command_with_arguments(
            command="config-set", service=self.service, arguments=config
        )

    def config_test(self, config: dict) -> KeaResponse:
        """Check whether the configuration supplied can be loaded by the dhcp4 daemon

        Args:
            config:     Configuration to test

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-test
        """
        return self.api.send_command_with_arguments(
            command="config-test", service=self.service, arguments=config
        )

    def config_write(self, filename: str) -> KeaResponse:
        """Write the current configuration to a file on disk

        Args:
            filename:       Name of the configuration file

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-write
        """
        return self.api.send_command_with_arguments(
            command="config-write",
            service=self.service,
            arguments={"filename": filename},
        )

    def dhcp_disable(self, max_period: int = 20) -> KeaResponse:
        """Globally disables DHCP service (dhcp4)

        Args:
            max_period:     Time until DHCP service is automatically renabled in seconds

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-dhcp-disable
        """
        return self.api.send_command_with_arguments(
            command="dhcp-disable",
            service=self.service,
            arguments={"max-period": max_period, "origin": "user"},
        )

    def dhcp_enable(self) -> KeaResponse:
        """Globally enables the DHCP service (dhcp4)

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/arm/ctrl-channel.html#the-dhcp-enable-command
        """
        return self.api.send_command_with_arguments(
            command="dhcp-enable", service=self.service, arguments={"origin": "user"}
        )

    def lease6_add(
        self, *, ip_address: str, duid: str, iaid: int, **kwargs
    ) -> KeaResponse:
        """Administratively add a new IPv6 lease

        Args:
            ip_address:         IPv6 Address of lease
            duid:               DHCP Unique Identifier
            iaid:               Identity Assosication Identifer

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-add
        """
        lease = Lease6(ip_address=ip_address, duid=duid, iaid=iaid, **kwargs)

        return self.api.send_command_with_arguments(
            command="lease6-add",
            service=self.service,
            arguments=lease.dict(exclude_none=True, by_alias=True),
            required_hook="lease_cmds",
        )

    def lease6_del(self, ip_address: str) -> KeaResponse:
        """Deletes a lease from the lease database

        Args:
            ip_address:     IP address of lease to delete

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-del
        """
        return self.api.send_command_with_arguments(
            command="lease6-del",
            service=self.service,
            arguments={"ip-address": ip_address},
            required_hook="lease_cmds",
        )

    def lease6_get(self, ip_address: str, type: Lease6TypeEnum = None) -> Lease6:
        """Queries the lease database and retrieves existing lease

        Args:
            ip_address:     IP address of lease

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease4-get
        """
        payload = {"ip-address": ip_address}
        if type:
            payload.update["type"] = type

        data = self.api.send_command_with_arguments(
            command="lease6-get",
            service=self.service,
            arguments=payload,
            required_hook="lease_cmds",
        )

        if data.result == 3:
            raise KeaLeaseNotFoundException(ip_address)

        return Lease6.parse_obj(data.arguments)

    def lease6_get_all(self, subnets: List[int] = []) -> List[Lease6]:
        """Retrieves all IPv6 leases or all leases for the specified subnets

        Args:
            subnets:        List of subnet IDs to fetch leases for

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-get-all
        """
        data = self.api.send_command_with_arguments(
            command="lease6-get-all",
            service=self.service,
            arguments={"subnets": subnets},
            required_hook="lease_cmds",
        )

        if data.result == 3:
            raise KeaLeaseNotFoundException(data.text)

        leases = [Lease6.parse_obj(lease) for lease in data.arguments["leases"]]
        return leases

    def lease6_get_by_duid(self, duid: str) -> Lease6:
        """Retrieves a lease for the specified duid

        Args:
            duid:       DHCP Unique Identifier

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-get-by-duid
        """
        data = self.api.send_command_with_arguments(
            command="lease6-get-by-duid",
            service=self.service,
            arguments={"duid": duid},
            required_hook="lease_cmds",
        )

        if data.result == 3:
            raise KeaLeaseNotFoundException(
                f"Unable to find a lease using duid '{duid}'"
            )

        lease = data.arguments["leases"][0]
        return Lease6.parse_obj(lease)

    def lease6_get_by_hostname(self, hostname: str) -> Lease6:
        """Retrieves all IPv6 leases for the specified hostname

        Args:
            hostname:   Hostname

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-get-by-hostname
        """
        data = self.api.send_command_with_arguments(
            command="lease6-get-by-hostname",
            service=self.service,
            arguments={"hostname": hostname},
            required_hook="lease_cmds",
        )

        if data.result == 3:
            raise KeaLeaseNotFoundException(
                f"Unable to find lease using hostname '{hostname}'"
            )

        lease = data.arguments["leases"][0]
        return Lease6.parse_obj(lease)

    def lease6_get_page(self, limit: int, search_from: str) -> Lease6Page:
        """Retrieves all IPv6 leases by page

        Args:
            limit:          Set the limit of IPv6 leases to be returned
            search_from:    Start from either a specific IP address or 'start' for the first

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-get-page
        """
        data = self.api.send_command_with_arguments(
            command="lease6-get-page",
            service=self.service,
            arguments={"from": search_from, "limit": limit},
            required_hook="lease_cmds",
        )

        return Lease6Page.parse_obj(data.arguments)

    def lease6_resend_ddns(self, ip_address: str) -> KeaResponse:
        """Sends an internal request to the ddns daemon to update DNS for an existing lease

        Args:
            ip_address:     Lease to update DDNS record

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease6-resend-ddns
        """
        return self.api.send_command_with_arguments(
            command="lease6-resend-ddns",
            service=self.service,
            arguments={"ip-address": ip_address},
            required_hook="lease_cmds",
        )

    def lease6_update(
        self, ip_address: str, duid: str, iaid: int, **kwargs
    ) -> KeaResponse:
        """Updates an existing lease

        Args:
            ip_address:     Lease IPv6 Address
            duid:           DHCP Unique Identifier
            iaid:           IAID

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#lease4-update
        """
        lease = Lease6(ip_address=ip_address, duid=duid, iaid=iaid, **kwargs)

        return self.api.send_command_with_arguments(
            command="lease6-update",
            service=self.service,
            arguments=lease.dict(exclude_none=True, by_alias=True),
            required_hook="lease_cmds",
        )

    def list_commands(self) -> KeaResponse:
        """List all commands supported by the server/service

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-list-commands
        """
        return self.api.send_command_with_arguments(
            command="list-commands", service=self.service, arguments={}
        )

    def network6_add(self, shared_networks: List[SharedNetwork6]) -> KeaResponse:
        """Adds new shared networks

        Args:
            shared_networks:        List of Shared Networks to add

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-network6-add
        """
        return self.api.send_command_with_arguments(
            command="network6-add",
            service=self.service,
            arguments={
                "shared-networks": [
                    network.dict(exclude_none=True, by_alias=True)
                    for network in shared_networks
                ]
            },
            required_hook="subnet_cmds",
        )

    def network6_del(self, name: str) -> KeaResponse:
        """Deletes an existing shared network

        Args:
            name:       Name of shared network

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#network6-del
        """
        return self.api.send_command_with_arguments(
            command="network6-del",
            service=self.service,
            arguments={"name": name},
            required_hook="subnet_cmds",
        )

    def network6_get(self, name: str) -> SharedNetwork6:
        """Returns detailed information about a shared network, including subnets

        Args:
            name:       Name of shared network

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#network6-get
        """
        data = self.api.send_command_with_arguments(
            command="network6-get",
            service=self.service,
            arguments={"name": name},
            required_hook="subnet_cmds",
        )

        if data.result == 3:
            raise KeaSharedNetworkNotFoundException(name)

        if not data.arguments["shared-networks"]:
            return None

        shared_network = data.arguments["shared-networks"][0]
        return SharedNetwork6.parse_obj(shared_network)

    def network6_list(self) -> List[SharedNetwork6]:
        """Returns a full list of the current shared networks configured

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#network6-list
        """
        data = self.api.send_command(
            command="network6-list",
            service=self.service,
            required_hook="subnet_cmds",
        )

        networks = [
            SharedNetwork6.parse_obj(network)
            for network in data.arguments["shared-networks"]
        ]
        return networks

    def network6_subnet_add(self, name: str, subnet_id: int) -> KeaResponse:
        """Add an existing subnet to an existing shared network

        Args:
            name:       Name of shared network
            subnet_id:  ID of the subnet

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#network6-subnet-add
        """
        return self.api.send_command_with_arguments(
            command="network6-subnet-add",
            service=self.service,
            arguments={"name": name, "id": subnet_id},
            required_hook="subnet_cmds",
        )

    def network6_subnet_del(self, name: str, subnet_id: int) -> KeaResponse:
        """Remove a subnet that is part of an existing shared network and demotes it to a plain standalone subnet

        Args:
            name:       Name of shared network
            subnet_id:  ID of the subnet

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#network6-subnet-del
        """
        return self.api.send_command_with_arguments(
            command="network6-subnet-del",
            service=self.service,
            arguments={"name": name, "id": subnet_id},
            required_hook="subnet_cmds",
        )

    def shutdown(self) -> KeaResponse:
        """Instructs the server daemon to initiate its shutdown procedure

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-shutdown
        """
        return self.api.send_command_with_arguments(
            command="shutdown", service=self.service, arguments={"exit-value": 3}
        )

    def statistic_get(self, name: str) -> KeaResponse:
        """Returns single statistic

        Args:
            name:       Name of the statistic to get

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-statistic-get
        """
        return self.api.send_command_with_arguments(
            command="statistic-get", service=self.service, arguments={"name": name}
        )

    def statistic_get_all(self) -> KeaResponse:
        """Returns all recorded statistics

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-statistic-get-all
        """
        return self.api.send_command_with_arguments(
            command="statistic-get-all", service=self.service, arguments={}
        )

    def status_get(self) -> StatusGet:
        """Returns servers runtime information

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-status-get
        """
        data = self.api.send_command(command="status-get", service=self.service)
        return StatusGet.parse_obj(data.arguments)

    def subnet6_add(self, subnets: List[Subnet6]) -> KeaResponse:
        """Creates and adds a new subnet

        Args:
            subnets:        List of subnets to add

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-add
        """
        return self.api.send_command_with_arguments(
            command="subnet6-add",
            service=self.service,
            arguments={
                "subnet6": [
                    subnet.dict(exclude_none=True, by_alias=True) for subnet in subnets
                ]
            },
            required_hook="subnet_cmds",
        )

    def subnet6_del(self, subnet_id: int) -> KeaResponse:
        """Removes a subnet

        Args:
            subnet_id:      ID of the subnet

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-del
        """
        data = self.api.send_command_with_arguments(
            command="subnet6-del",
            service=self.service,
            arguments={"id": subnet_id},
            required_hook="subnet_cmds",
        )

        if data.result == 3:
            raise KeaSubnetNotFoundException(subnet_id)

        return data

    def subnet6_delta_add(self, subnets: List[Subnet6]) -> KeaResponse:
        """Updates (adds or overwrites) parts of a single subnet

        Args:
            subnets:        List of subnets to update/add

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-delta-add
        """
        return self.api.send_command_with_arguments(
            command="subnet6-delta-add",
            service=self.service,
            arguments={
                "subnet6": [
                    subnet.dict(exclude_none=True, by_alias=True) for subnet in subnets
                ]
            },
            required_hook="subnet_cmds",
        )

    def subnet6_delta_del(self, subnets: List[Subnet6]) -> KeaResponse:
        """Updates (removes) parts of a single subnet

        Args:
            subnets:        List of subnets to update/delete

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-delta-del
        """
        return self.api.send_command_with_arguments(
            command="subnet6-delta-del",
            service=self.service,
            arguments={
                "subnet6": [
                    subnet.dict(exclude_none=True, by_alias=True) for subnet in subnets
                ]
            },
            required_hook="subnet_cmds",
        )

    def subnet6_get(self, subnet_id: int) -> Subnet6:
        """Gets detailed information about the specified subnet

        Args:
            subnet_id:      ID of the subnet

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-get
        """
        data = self.api.send_command_with_arguments(
            command="subnet6-get",
            service=self.service,
            arguments={"id": subnet_id},
            required_hook="subnet_cmds",
        )

        if data.result == 3:
            raise KeaSubnetNotFoundException(subnet_id)

        if not data.arguments["subnet6"]:
            return None

        subnet = data.arguments["subnet6"][0]
        return Subnet6.parse_obj(subnet)

    def subnet6_list(self) -> List[Subnet6]:
        """List all currently configured subnets

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-list
        """
        data = self.api.send_command(
            command="subnet6-list",
            service=self.service,
            required_hook="subnet_cmds",
        )

        subnets = [Subnet6.parse_obj(subnet) for subnet in data.arguments["subnets"]]
        return subnets

    def subnet6_update(self, subnets: List[Subnet6]) -> List[Subnet6]:
        """Updates (overwrites) a single subnet

        Args:
            subnets:        List of subnets to overwrite

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#subnet6-update
        """
        return self.api.send_command_with_arguments(
            command="subnet6-update",
            service=self.service,
            arguments={
                "subnet6": [
                    subnet.dict(exclude_none=True, by_alias=True) for subnet in subnets
                ]
            },
            required_hook="subnet_cmds",
        )

    def version_get(self) -> KeaResponse:
        """Returns extended information about the Kea Version that is running

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-version-get
        """
        return self.api.send_command(command="version-get", service=self.service)



class KeaException(Exception):
    def __init__(self, message: str = "Kea Exception encountered"):
        self.message = message
        super().__init__(self.message)


class KeaGenericException(KeaException):
    # Result code 1
    def __init__(
        self,
        message: str = "Code 1: a general error or failure has occurred during the command processing.",
    ):
        self.message = message
        super().__init__(self.message)


class KeaCommandNotSupportedException(KeaException):
    # Result code 2
    def __init__(
        self,
        message: str = "Code 2: the specified command is unsupported by the server receiving it.",
    ):
        self.message = message
        super().__init__(self.message)


class KeaObjectNotFoundException(KeaException):
    # Result code 3
    def __init__(
        self,
        message: str = "Code 3: the requested operation has been completed but the requested resource was not found. This status code is returned when a command returns no resources or affects no resources.",
    ):
        self.message = message
        super().__init__(self.message)


class KeaServerConflictException(KeaException):
    # Result code 4
    def __init__(
        self,
        message: str = "Code 4: the well-formed command has been processed but the requested changes could not be applied, because they were in conflict with the server state or its notion of the configuration.",
    ):
        self.message = message
        super().__init__(self.message)


class KeaUnauthorizedAccessException(KeaException):
    def __init__(
        self,
        message: str = "Received 401 (Unauthorized response), please ensure username and password is configured, and use_basic_auth is set to True for the Kea object",
    ):
        self.message = message
        super().__init__(self.message)


class KeaHookLibraryNotConfiguredException(KeaException):
    def __init__(self, service: str, hook: str):
        self.message = f"Hook library '{hook}' is not configured for '{service}' service. Please ensure this is enabled in the configuration for the '{service}' daemon"
        super().__init__(self.message)


class KeaSharedNetworkNotFoundException(KeaException):
    def __init__(self, name: str):
        self.message = f"Shared Network '{name}' not found"
        super().__init__(self.message)


class KeaSubnetNotFoundException(KeaException):
    def __init__(self, subnet_id: int):
        self.message = f"Subnet with id '{subnet_id}' not found"
        super().__init__(self.message)


class KeaLeaseNotFoundException(KeaException):
    def __init__(self, lease: str):
        self.message = f"Lease '{lease}' not found"
        super().__init__(self.message)



from typing import Optional, List
from pykeadhcp.models.generic.daemon import CommonDhcpDaemonConfig
from pykeadhcp.models.dhcp6.client_class import ClientClass6
from pykeadhcp.models.dhcp6.shared_network import SharedNetwork6
from pykeadhcp.models.dhcp6.subnet import Subnet6
from pykeadhcp.models.dhcp6.reservation import Reservation6
from pykeadhcp.models.dhcp6.server_id import ServerId


class Dhcp6DaemonConfig(CommonDhcpDaemonConfig):
    client_classes: Optional[List[ClientClass6]]
    shared_networks: Optional[List[SharedNetwork6]]
    reservations: Optional[List[Reservation6]]
    data_directory: Optional[str]
    preferred_lifetime: Optional[int]
    min_preferred_lifetime: Optional[int]
    max_preferred_lifetime: Optional[int]
    subnet6: Optional[List[Subnet6]]
    mac_sources: Optional[List[str]]
    relay_supplied_options: Optional[List[str]]
    server_id: ServerId



from pydantic import BaseModel


def normalize_keys(string: str) -> str:
    return string.replace("_", "-")


class KeaBaseModel(BaseModel):
    class Config:
        alias_generator = normalize_keys
        allow_population_by_field_name = True
        use_enum_values = True



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pykeadhcp import Kea

from pykeadhcp.models.generic import KeaResponse


class Ddns:
    def __init__(self, api: "Kea"):
        self.service = "Ddns"
        self.api = api

        # Cache config and hooks
        try:
            self.cached_config = None
            self.refresh_cached_config()
            self.hook_libraries = self.api.get_active_hooks(
                hooks=self.cached_config[self.service.capitalize()]["hooks-libraries"]
            )
            self.api.hook_library[self.service] = self.hook_libraries
        except:
            pass

    def refresh_cached_config(self):
        """Sets the cached_config variable

        This function should be called after any interaction with the API that potentially changes the configuration
        eg. config-set, commands like config-test won't need a config refresh to keep the cached config up to date
        """
        config = self.config_get()
        self.cached_config = config.arguments

    def build_report(self) -> KeaResponse:
        """Returns list of compilation options that this particular binary was built with

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-build-report
        """
        return self.api.send_command(command="build-report", service=self.service)

    def config_get(self) -> KeaResponse:
        """Retrieves the current configuration used by the server

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-config-get
        """
        return self.api.send_command(command="config-get", service=self.service)

    def config_reload(self) -> KeaResponse:
        """Reloads the last good configuration (configuration file on disk)

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-config-reload
        """
        return self.api.send_command(command="config-reload", service=self.service)

    def config_set(self, config: dict) -> KeaResponse:
        """Replace the current server configuration with the provided configuration

        Args:
            config:     Configuration to set

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-set
        """
        return self.api.send_command_with_arguments(
            command="config-set", service=self.service, arguments=config
        )

    def config_test(self, config: dict) -> KeaResponse:
        """Check whether the configuration supplied can be loaded by the dhcp4 daemon

        Args:
            config:     Configuration to test

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-test
        """
        return self.api.send_command_with_arguments(
            command="config-test", service=self.service, arguments=config
        )

    def config_write(self, filename: str) -> KeaResponse:
        """Write the current configuration to a file on disk

        Args:
            filename:       Name of the configuration file

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#config-write
        """
        return self.api.send_command_with_arguments(
            command="config-write",
            service=self.service,
            arguments={"filename": filename},
        )

    def list_commands(self) -> KeaResponse:
        """List all commands supported by the server/service

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-list-commands
        """
        return self.api.send_command_with_arguments(
            command="list-commands", service=self.service, arguments={}
        )

    def statistic_get(self, name: str) -> KeaResponse:
        """Returns single statistic

        Args:
            name:       Name of the statistic to get

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-statistic-get
        """
        return self.api.send_command_with_arguments(
            command="statistic-get", service=self.service, arguments={"name": name}
        )

    def statistic_get_all(self) -> KeaResponse:
        """Returns all recorded statistics

        Kea API Reference:
            https://kea.readthedocs.io/en/kea-2.2.0/api.html#ref-statistic-get-all
        """
        return self.api.send_command_with_arguments(
            command="statistic-get-all", service=self.service, arguments={}
        )



from typing import Optional
from pykeadhcp.models.generic.config import CommonConfig
from pykeadhcp.models.enums import ServerIdTypeEnum


class ServerId(CommonConfig):
    type: ServerIdTypeEnum
    htype: int
    identifier: Optional[str]
    time: int
    enterprise_id: int
    persist: bool



from typing import Optional
from pykeadhcp.models.generic.base import KeaBaseModel


class Lease(KeaBaseModel):
    cltt: Optional[int]
    fqdn_fwd: Optional[bool]
    fqdn_rev: Optional[bool]
    hostname: Optional[str]
    hw_address: Optional[str]
    ip_address: str
    state: Optional[int]
    subnet_id: Optional[int]
    valid_lft: Optional[int]


class LeasePage(KeaBaseModel):
    count: int



from typing import Optional, List
from pykeadhcp.models.generic.lease import Lease, LeasePage
from pykeadhcp.models.enums import Lease6TypeEnum


class Lease6(Lease):
    duid: str
    iaid: int
    prefix_len: Optional[int]
    type: Optional[Lease6TypeEnum]


class Lease6Page(LeasePage):
    count: int
    leases: Optional[List[Lease6]]



from pykeadhcp.parsers.generic import GenericParser
from pykeadhcp.models.ctrlagent.config import CtrlAgentDaemonConfig


class CtrlAgentParser(GenericParser):
    """Parser for the ISC Kea CtrlAgent configuration file. This should ideally
    be used with the cached config stored in the Daemon class like this:

    parser = CtrlAgentParser(config=server.ctrlagent.cached_config)
    """

    def __init__(self, config: dict):
        self.config = CtrlAgentDaemonConfig.parse_obj(config["Control-agent"])



from typing import Optional, List
from pydantic import conint
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.enums import LoggerLevelEnum


class Output(KeaBaseModel):
    output: str
    flush: bool
    maxsize: int
    maxver: int
    pattern: Optional[str]


class Logger(KeaBaseModel):
    name: str
    output_options: Optional[List[Output]]
    debuglevel: conint(ge=0, le=100)
    severity: Optional[LoggerLevelEnum]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]

    class Config:
        fields = {"output_options": "output_options"}



from typing import Optional, List
from pykeadhcp.models.generic.config import CommonConfig
from pykeadhcp.models.enums import AuthenticationTypeEnum


class AuthenticationClient(CommonConfig):
    user: Optional[str]
    user_file: Optional[str]
    password: Optional[str]
    password_file: Optional[str]


class Authentication(CommonConfig):
    type: AuthenticationTypeEnum
    realm: str
    directory: Optional[str]
    clients: List[Optional[AuthenticationClient]]



from typing import Optional, List
from pykeadhcp.models.generic.base import KeaBaseModel
from pykeadhcp.models.generic.option_data import OptionData


class Reservation(KeaBaseModel):
    duid: Optional[str]
    client_classes: Optional[List[str]]
    flex_id: Optional[str]
    hw_address: Optional[str]
    hostname: Optional[str]
    option_data: Optional[List[OptionData]]
    user_context: Optional[dict]
    comment: Optional[str]
    unknown_map_entry: Optional[str]



from typing import List, Optional
from pykeadhcp.models.generic.lease import Lease, LeasePage


class Lease4(Lease):
    pass


class Lease4Page(LeasePage):
    count: int
    leases: Optional[List[Lease4]]
